//----------------------------------------------------------------------------
// union-find framework for random maze generation
// copyright 2010 christopher rasmussen
// university of delaware
//----------------------------------------------------------------------------

#include <iostream>
#include <ctype.h>
#include <fstream>
#include <vector>
#include <cstdlib>
#include "sys/time.h"

using namespace std;

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------

class DisjointSet
{
public:

  DisjointSet(int);
  void print();
  void merge(int, int);
  int find(int);
  bool same_equivalence_class(int x, int y) { return find(x) == find(y); }

  vector <int> S;   // forest of trees
  int num_equivalence_classes;
  
};

//----------------------------------------------------------------------------

DisjointSet::DisjointSet(int num_elements)
{
  S.clear();
  S.resize(num_elements);

  for (int i = 0; i < S.size(); i++)
    S[i] = -1;

  num_equivalence_classes = S.size();
}

//----------------------------------------------------------------------------

void DisjointSet::print()
{
  cout << num_equivalence_classes << " equivalence classes\n";

  for (int i = 0; i < S.size(); i++) {
    if (S[i] < 0)
      cout << "* ";
    else
      cout << "  ";
    cout << i << ": " << S[i] << endl;
  }
}

//----------------------------------------------------------------------------

// "dumb" union

void DisjointSet::merge(int x, int y)
{
  int root_x = find(x);
  int root_y = find(y);

  if (root_x != root_y) {
    S[root_y] = root_x;
    num_equivalence_classes--;
  }
}

//----------------------------------------------------------------------------

// includes path correction

int DisjointSet::find(int x)
{
  if (x < 0 || x >= S.size()) {
    cout << "find(): " << x << " is not a set element\n";
    exit(1);
  }

  if (S[x] < 0) 
    return x;
  else
    return (S[x] = find(S[x]));
}

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------

class Maze
{
public:

  Maze(int, int);
  void print();

  // is the wall BELOW the cell at (row, col) an exterior one?
  bool is_exterior_bottom_wall(int row, int col) { return (row == num_rows - 1); }

  // is the wall to the LEFT of the cell at (row, col) an exterior one?
  bool is_exterior_left_wall(int row, int col) { return (col == 0); }

  // convert grid cell to set element index
  int element_index(int row, int col)  { return (col + row * num_cols); }

  void knock_down_wall(int, int, bool);
  bool pick_random_interior_wall(int &, int &, bool &);
  void knock_down_all_walls_randomly(bool);

  void knock_down_walls_randomly_until_path_exists(bool);
  void knock_down_walls_randomly_until_everything_connected(bool);

  int num_rows, num_cols;

  int start_row, start_col, start_index;
  int finish_row, finish_col, finish_index;

  int num_interior_walls;
  int num_knockdowns;
  vector < vector <bool> > bottom_wall;
  vector < vector <bool> > left_wall;

  DisjointSet *S;
  
};

//----------------------------------------------------------------------------

Maze::Maze(int rows, int cols)
{
  // initialize random number generator

  srand(time(0));

  num_rows = rows;
  num_cols = cols;

  num_interior_walls = (num_rows - 1) * num_cols + (num_cols - 1) * num_rows;
  num_knockdowns = 0;

  start_row = start_col = 0;  // upper-left by default
  start_index = element_index(start_row, start_col);

  finish_row = num_rows - 1;  // lower-right by default
  finish_col = num_cols - 1;
  finish_index = element_index(finish_row, finish_col);

  // initialize all walls 

  bottom_wall.resize(num_rows);
  left_wall.resize(num_rows);

  for (int r = 0; r < num_rows; r++) {

    bottom_wall[r].resize(num_cols);
    left_wall[r].resize(num_cols);

    for (int c = 0; c < num_cols; c++) {
      bottom_wall[r][c] = true;
      left_wall[r][c] = true;
    }
  }

  // initialize underlying set data structure

  S = new DisjointSet(num_rows * num_cols);
}

//----------------------------------------------------------------------------

// remove wall at (row, col) and merge cells that it divides
// if bottom is false, knock down left wall

void Maze::knock_down_wall(int row, int col, bool bottom)
{
  int x, y;

  x = element_index(row, col);

  if (bottom) {
    bottom_wall[row][col] = false;
    y = element_index(row + 1, col);
  }
  else {
    left_wall[row][col] = false;
    y = element_index(row, col - 1);
  }

  S->merge(x, y);
  num_interior_walls--;
  num_knockdowns++;
}

//----------------------------------------------------------------------------

// brute force approach: pick random cell, pick random direction
// (bottom or left), check if wall is actually there.  repeat until
// one is found

bool Maze::pick_random_interior_wall(int & row, int & col, bool & bottom)
{
  if (num_interior_walls == 0)
    return false;
  else {
    while (1) {
      row = rand() % num_rows;
      col = rand() % num_cols;
      bottom = rand() % 2;
      if (bottom && !is_exterior_bottom_wall(row, col) && bottom_wall[row][col])
	break;
      else if (!bottom && !is_exterior_left_wall(row, col) && left_wall[row][col])
	break;
    }
    return true;
  }
}

//----------------------------------------------------------------------------

// print maze including current state of walls

void Maze::print()
{
  int r, c;

  cout << "  S";  // assuming upper-left start location 
  for (c = 0; c < num_cols - 1; c++)
    cout << " _";
  cout << endl;

  for (r = 0; r < num_rows; r++) {
    cout << r % 10 << "";
    for (c = 0; c < num_cols; c++) {
      
      if (left_wall[r][c])
	cout << "|";
      else
	cout << " ";

      if (bottom_wall[r][c])
	cout << "_";
      else
	cout << " ";

    }
    if (r < num_rows - 1)
      cout << "|\n";
    else
      cout << " F\n";
  }

  cout << " ";
  for (c = 0; c < num_cols; c++)
    cout << " " << c % 10;
  cout << "\n";

  cout << "num interior walls left = " << num_interior_walls << endl;
  cout << "num walls knocked down = " << num_knockdowns << endl;
}

//----------------------------------------------------------------------------

void Maze::knock_down_all_walls_randomly(bool interactive)
{
  int r, c;
  bool bottom;

  while (1) {

    // print current state of maze

    print();

    // get keyboard input (enter only will do)

    if (interactive)
      cin.get();

    // pick a wall (if there is one) and knock it down

    if (pick_random_interior_wall(r, c, bottom)) 
      knock_down_wall(r, c, bottom);
    else 
      break;
  }
}

//----------------------------------------------------------------------------

// FILL THIS IN!!!

void knock_down_walls_randomly_until_path_exists(bool interactive)
{
  // ???
}

//----------------------------------------------------------------------------

// FILL THIS IN!!!

void knock_down_walls_randomly_until_everything_connected(bool interactive)
{
  // ???
}

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------

int main(int argc, char **argv)
{
  Maze M1(3, 4);
  
  M1.knock_down_all_walls_randomly(true);
//   M1.knock_down_walls_randomly_until_path_exists(true);
//   M1.knock_down_walls_randomly_until_everything_connected(true);

//  Maze M2(15, 20);
  
  // ...

  return 1;
}
